@@ -255,7 +255,9 @@ void udf_init()
   if (unlikely(error > 0))
     sql_print_error("Got unknown error: %d", my_errno);
   end_read_record(&read_record_info);
-  table->m_needs_reopen= TRUE;                  // Force close to free memory
+
+  // Force close to free memory
+  table->mark_table_for_reopen();
 
 end:
   close_mysql_tables(new_thd);
