@@ -13,7 +13,7 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335 USA */
 
-#include <my_global.h>
+#include "mariadb.h"
 #include <violite.h>
 #include <sql_priv.h>
 #include <sql_class.h>
@@ -63,9 +63,6 @@ typedef OVERLAPPED_ENTRY native_event;
 #error threadpool is not available on this platform
 #endif
 
-#ifdef _MSC_VER
-#pragma warning (disable : 4312)
-#endif
 
 static void io_poll_close(TP_file_handle fd)
 {
@@ -181,7 +178,7 @@ connection_queue_t;
 
 const int NQUEUES=2; /* We have high and low priority queues*/
 
-struct thread_group_t 
+struct MY_ALIGNED(CPU_LEVEL1_DCACHE_LINESIZE) thread_group_t
 {
   mysql_mutex_t mutex;
   connection_queue_t queues[NQUEUES];
@@ -199,7 +196,7 @@ struct thread_group_t
   int  shutdown_pipe[2];
   bool shutdown;
   bool stalled; 
-} MY_ALIGNED(CPU_LEVEL1_DCACHE_LINESIZE);
+};
 
 static thread_group_t *all_groups;
 static uint group_count;
@@ -594,11 +591,8 @@ static void timeout_check(pool_timer_t *timer)
   THD *thd;
   while ((thd=it++))
   {
-    if (thd->net.reading_or_writing != 1)
-      continue;
- 
     TP_connection_generic *connection= (TP_connection_generic *)thd->event_scheduler.data;
-    if (!connection)
+    if (!connection || connection->state != TP_STATE_IDLE)
     {
       /* 
         Connection does not have scheduler data. This happens for example
@@ -1084,10 +1078,7 @@ void thread_group_destroy(thread_group_t *thread_group)
 #endif
 
   if (my_atomic_add32(&shutdown_group_count, -1) == 1)
-  {
     my_free(all_groups);
-    all_groups= 0;
-  }
 }
 
 /**
@@ -1505,7 +1496,7 @@ void TP_connection_generic::set_io_timeout(int timeout_sec)
 }
 
 
-
+#ifndef HAVE_IOCP
 /**
   Handle a (rare) special case,where connection needs to 
   migrate to a different group because group_count has changed
@@ -1540,7 +1531,7 @@ static int change_group(TP_connection_generic *c,
   mysql_mutex_unlock(&new_group->mutex);
   return ret;
 }
-
+#endif
 
 int TP_connection_generic::start_io()
 {
@@ -1680,14 +1671,6 @@ TP_pool_generic::~TP_pool_generic()
   {
     thread_group_close(&all_groups[i]);
   }
-
-  /*
-    Wait until memory occupied by all_groups is freed.
-  */
-  int timeout_ms=5000;
-  while(all_groups && timeout_ms--)
-    my_sleep(1000);
-
   threadpool_started= false;
   DBUG_VOID_RETURN;
 }
